#include "tcpip.h"

struct realm *tcp_realm;

struct tcp_head *tcp_rules=NULL;
struct tcp_head *tcp_rules_stateful=NULL;
unsigned int num_tcp_head=0;
unsigned int num_tcp_rules=0;

char *tcp_ignore_words[]={
	"stateless",
	"flow",
	NULL
};

/* Generator */
struct generator tcp_gen=init_generator("sig.tcp");

void tcp_match(struct packet *p, unsigned int l)
{
	struct tcp_head *h;
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	unsigned int i=0;
	u_int16_t sp, dp;

	iph=p->layer[l].h.ip;
	tcph=p->layer[l+1].h.tcp;

	sp=ntohs(tcph->sport);
	dp=ntohs(tcph->dport);

	h=tcp_rules;

tcp_iter:
	for(; h; h=h->next) {
		if ( 
			(((h->saddr&h->smask) == (iph->saddr&h->smask))
			^ ((h->flags&NEGATE_SADDR)!=0) ) &&
			
			(((h->daddr&h->dmask) == (iph->daddr&h->dmask))
			^ ((h->flags&NEGATE_DADDR)!=0) ) &&

			(((sp>=h->sport_min)&&(sp<=h->sport_max)))
			^ ((h->flags&NEGATE_SPORT)!=0) &&
			
			(((dp>=h->dport_min)&&(dp<=h->dport_max)))
			^ ((h->flags&NEGATE_DPORT)!=0)){

			if (tcpip_traverse(&tcp_gen, p, l, h->rules))
				return;
		}
	}

	/* If we are doing state tracking, and the state
	 * bit isn't set, then we don't match against
	 * the stateful rules! */
	if ( i!=0 || ((p->layer[l].flags&FLAG_TCP_TRACK) &&
		!(p->layer[l].flags&FLAG_TCP_STATE)) ) {
		return;
	}

	h=tcp_rules_stateful;
	i++;
	goto tcp_iter;
}

int tcp_commit(struct rule *r)
{
	struct tcp_head **tcph=&tcp_rules_stateful;
	struct criteria *c;

	/* Stateful by default. */
	for(c=r->criteria; c; c=c->next) {
		if ( !strcmp("stateless", c->crit) ) {
			tcph=&tcp_rules;
		}
	}

	return tcp_udp_commit(tcph,
		&num_tcp_head,
		&num_tcp_rules, r,
		tcp_ignore_words);
}

struct proto tcp_p=init_proto_ex("tcp", tcp_decode, tcp_dprint);
struct proto_req tcp_r[]={
	proto_request("ipv4", 6),
	null_request()
};

int tcp_dprint(struct layer *l, char *buf, int buflen)
{
	char tcpflags[]="FSRPAU12";
	char disp[9];
	int n, x;
	
	for(n=0,x=1; n<8; n++, x*=2) {
		if ( l->h.tcp->flags.flags & x ){
			disp[n]=tcpflags[n];
		}else{
			disp[n]='*';
		}
	}disp[8]='\0';
	
	return snprintf(buf, buflen,
		"%u > %u [%s] SEQ=0x%x ACK=0x%x WIN=0x%x",
		ntohs(l->h.tcp->sport),
		ntohs(l->h.tcp->dport),
		disp,
		ntohl(l->h.tcp->seq),
		ntohl(l->h.tcp->ack),
		ntohs(l->h.tcp->win));
}

int tcp_decode(struct packet *p)
{
	struct proto_child *pc;
	struct layer *l=&p->layer[p->llen-1];
	int ret=0;

	if ( p->llen >= PKT_LAYERS ) return ret;

	/* Check it is at least minimum size */
	if ( l->h.raw+sizeof(struct pkt_tcphdr) >= p->end )
		return ret;
	
	/* Use the doff field to decode the packet */
	if ( (p->layer[p->llen].h.raw=
		l->h.raw+(l->h.tcp->doff<<2))
		>= p->end ) {
		p->layer[p->llen].h.raw=p->end;
		return ret;
	}

	for(pc=l->proto->children; pc; pc=pc->next)
	{
		if ( l->h.tcp->dport == pc->id ||
			l->h.tcp->sport == pc->id ) {
			p->layer[p->llen].flags=0;
			p->layer[p->llen].session=NULL;
			p->layer[p->llen++].proto=pc->proto;
			return pc->proto->decode(p) | ret;
		}
	}

	p->layer[p->llen++].proto=NULL;
	return ret;
}
