/* Copyright (c) 2002 Gianni Tedesco <gianni@ecsc.co.uk>
 *
 * Some code taken from libnids:
 * Copyright (c) 1999 Rafal Wojtczuk <nergal@avet.com.pl>. All rights reserved.
 *
 * NOTE: This is dreadful, we need a more advanced implementation.
*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <packet.h>
#include <preproc.h>
#include <decode.h>
#include <plugin.h>

#if 1
#define dprintf(x,...)
#define __INLINE__ static inline
#else
#define dprintf printf
#define __INLINE__
#endif

/* Hacky shit to get rid of */
#define b_comp(x,y) (!memcmp(&(x), &(y), sizeof(x)))

__INLINE__ int after(u_int seq1, u_int seq2){return ((int)(seq2 - seq1) < 0);}
__INLINE__ int before(u_int seq1, u_int seq2){return ((int)(seq1 - seq2) < 0);}

enum
{
	TCP_ESTABLISHED = 1,
	TCP_SYN_SENT,
	TCP_SYN_RECV,
	TCP_FIN_WAIT1,
	TCP_FIN_WAIT2,
	TCP_TIME_WAIT,
	TCP_CLOSE,
	TCP_CLOSE_WAIT,
	TCP_LAST_ACK,
	TCP_LISTEN,
	TCP_CLOSING,

	/* Non-standard states */
	FIN_SENT,
	FIN_CONFIRMED
};

#define NIDS_JUST_EST 1
#define NIDS_DATA 2
#define NIDS_CLOSE 3
#define NIDS_RESET 4
#define NIDS_TIMED_OUT 5

struct tuple4 {
	u_int16_t source;
	u_int16_t dest;
	u_int32_t saddr;
	u_int32_t daddr;
};

struct half_stream {
	char state;
	unsigned int seq;
	unsigned int ack_seq;
	unsigned int first_data_seq;
	unsigned short window;
};

struct tcp_stream {
	struct tuple4 addr;
	char nids_state;
	struct half_stream client;
	struct half_stream server;
	struct tcp_stream *next_node;
	struct tcp_stream *prev_node;
	int hash_index;
	struct tcp_stream *next_time;
	struct tcp_stream *prev_time;
	struct tcp_stream *next_free;
};

/* Hash table and various lists */
struct tcp_stream **tcp_stream_table;
struct tcp_stream *streams_pool;
struct tcp_stream *tcp_latest = 0, *tcp_oldest = 0;
struct tcp_stream *free_streams;

/* Statistics */
int tcp_num=0;
int max_stream;
unsigned int tcp_packets=0;
unsigned int tcp_state_errors=0;

/* Stuff we import */
proc_preproc_activate preproc_activate;
proc_preproc_dispatch preproc_dispatch;
proc_decode_proto decode_proto;
struct proto *tcpproto=NULL;

/* Mask to determine fragments */
unsigned short ipfmask=__constant_htons(IP_MF|IP_OFFMASK);

/* Must be a power of 2 */
#define TCPHASH 512

/* TCP hashing function */
__INLINE__ unsigned int tcp_hashfn(
	u_int32_t saddr, u_int32_t daddr, 
	u_int16_t sport, u_int16_t dport)
{
	unsigned int h = ((daddr ^ dport) ^ (saddr ^ sport));
	h ^= h>>16;
	h ^= h>>8;
	return h & (TCPHASH-1);
}

static void free_tcp (struct tcp_stream *a_tcp)
{
	int hash_index = a_tcp->hash_index;

	if (a_tcp->next_node)
		a_tcp->next_node->prev_node = a_tcp->prev_node;
	if (a_tcp->prev_node)
		a_tcp->prev_node->next_node = a_tcp->next_node;
	else
		tcp_stream_table[hash_index] = a_tcp->next_node;
	if (a_tcp->next_time)
		a_tcp->next_time->prev_time = a_tcp->prev_time;
	if (a_tcp->prev_time)
		a_tcp->prev_time->next_time = a_tcp->next_time;
	if (a_tcp == tcp_oldest)
		tcp_oldest = a_tcp->prev_time;
	if (a_tcp == tcp_latest)
		tcp_latest = a_tcp->next_time;

	a_tcp->next_free = free_streams;
	free_streams = a_tcp;
	tcp_num--;
}

__INLINE__ unsigned int mk_hash_index (struct tuple4 addr)
{
	return tcp_hashfn(addr.saddr, addr.daddr, addr.source, addr.dest);
}

struct tcp_stream *add_new_tcp(struct pkt_tcphdr *this_tcphdr,
	struct pkt_iphdr *this_iphdr)
{
	struct tcp_stream *tolink;
	struct tcp_stream *a_tcp;
	int hash_index;
	struct tuple4 addr;

	addr.source = ntohs (this_tcphdr->sport);
	addr.dest = ntohs (this_tcphdr->dport);
	addr.saddr = this_iphdr->saddr;
	addr.daddr = this_iphdr->daddr;
	hash_index = mk_hash_index (addr);

	if (tcp_num > max_stream) {
		/* NOTIFY: Timeout */
		dprintf("tcpstream: timed out\n");
		tcp_oldest->nids_state = NIDS_TIMED_OUT;
		free_tcp(tcp_oldest);
	}
	
	if ( !(a_tcp=free_streams) ) {
		return NULL;
	}

	tcp_num++;
	free_streams = a_tcp->next_free;
	tolink = tcp_stream_table[hash_index];
	memset (a_tcp, 0, sizeof (struct tcp_stream));
	a_tcp->hash_index = hash_index;
	a_tcp->addr = addr;
	a_tcp->client.state = TCP_SYN_SENT;
	a_tcp->client.seq = ntohl (this_tcphdr->seq) + 1;
	a_tcp->client.first_data_seq = a_tcp->client.seq;
	a_tcp->client.window = ntohs (this_tcphdr->win);
	a_tcp->server.state = TCP_CLOSE;
	a_tcp->next_node = tolink;
	a_tcp->prev_node = 0;
	if (tolink) tolink->prev_node = a_tcp;
	tcp_stream_table[hash_index] = a_tcp;
	a_tcp->next_time = tcp_latest;
	a_tcp->prev_time = 0;
	if (!tcp_oldest) tcp_oldest = a_tcp;
	if (tcp_latest) tcp_latest->prev_time = a_tcp;
	tcp_latest = a_tcp;

	return a_tcp;
}

struct tcp_stream *find_stream (struct pkt_tcphdr *this_tcphdr,
				struct pkt_iphdr *this_iphdr, int *from_client)
{
	struct tuple4 this_addr, reversed;
	int hash_index;
	struct tcp_stream *a_tcp;

	/* First one way */
	this_addr.source = ntohs (this_tcphdr->sport);
	this_addr.dest = ntohs (this_tcphdr->dport);
	this_addr.saddr = this_iphdr->saddr;
	this_addr.daddr = this_iphdr->daddr;
	hash_index = mk_hash_index (this_addr);
	for (a_tcp=tcp_stream_table[hash_index];
		a_tcp; a_tcp = a_tcp->next_node) {
		if ( b_comp(a_tcp->addr, this_addr) ) {
			*from_client = 1;
			return a_tcp;
		}
	}
	
	/* Then the other */
	reversed.source = ntohs (this_tcphdr->dport);
	reversed.dest = ntohs (this_tcphdr->sport);
	reversed.saddr = this_iphdr->daddr;
	reversed.daddr = this_iphdr->saddr;
	hash_index = mk_hash_index (reversed);
	for (a_tcp = tcp_stream_table[hash_index];
	     a_tcp ; a_tcp = a_tcp->next_node) {
		if ( b_comp(a_tcp->addr, reversed) ) {
			*from_client = 0;
			return a_tcp;
		}
	}

	return NULL;
}

unsigned int process_tcp (struct packet *p, unsigned int i)
{
	struct pkt_iphdr *this_iphdr=p->layer[i].h.ip;
	struct pkt_tcphdr *this_tcphdr=p->layer[i+1].h.tcp;
	int datalen;
	int from_client = 1;
	struct tcp_stream *a_tcp;
	struct half_stream *snd, *rcv;
	unsigned int ret=0;
	
	datalen = ntohs(this_iphdr->tot_len) - 
		(this_iphdr->ihl<<2) - 
		(this_tcphdr->doff<<2);

	/* TODO: Check TCP checksum */

	/* Try to find the corresponding stream for this packet */
	if (!(a_tcp=find_stream(this_tcphdr, this_iphdr, &from_client)))
	{
		/* If one isn't found this could be the
		 * first packet of an active open */
		if ((this_tcphdr->flags.bits.syn ) &&
		    !(this_tcphdr->flags.bits.ack) &&
		    !(this_tcphdr->flags.bits.rst)) {
			p->layer[i].session=add_new_tcp(this_tcphdr, this_iphdr);
			return FLAG_TCP_STATE;
		}
		return 0;
	}
	
	if (from_client){
		snd = &a_tcp->client;
		rcv = &a_tcp->server;
		ret|= FLAG_TCP_STATE|FLAG_TCP_2SVR;
	}else{
		rcv = &a_tcp->client;
		snd = &a_tcp->server;
		ret|= FLAG_TCP_STATE;
	}

	/* SYN packets */
	if (this_tcphdr->flags.bits.syn)
	{
		/* Out of state */
		if (from_client || a_tcp->client.state != TCP_SYN_SENT ||
		    a_tcp->server.state != TCP_CLOSE
		    || !(this_tcphdr->flags.bits.ack))
			return 0;
		
		/* Spoofing attack */
		if (a_tcp->client.seq != ntohl (this_tcphdr->ack))
			return 0;
		
		/* Successful SYN */
		a_tcp->server.state = TCP_SYN_RECV;
		a_tcp->server.seq = ntohl (this_tcphdr->seq) + 1;
		a_tcp->server.first_data_seq = a_tcp->server.seq;
		a_tcp->server.ack_seq = ntohl (this_tcphdr->ack);
		a_tcp->server.window = ntohs (this_tcphdr->win);
		return ret;
	}
	
	/* Window tracking - All beyond this point is ASSURED */
	if (!before (ntohl (this_tcphdr->seq), rcv->ack_seq + rcv->window)
	    || before (ntohl (this_tcphdr->seq) + datalen, rcv->ack_seq)){
		return 0;
	}else{
		ret|= FLAG_TCP_SURE;
	}

	/* Deal with RST packets */
	if (this_tcphdr->flags.bits.rst)
	{
		if (a_tcp->nids_state == NIDS_DATA)
		{
			a_tcp->nids_state = NIDS_RESET;
			/* NOTIFY: Reset by RST */
			dprintf("tcpstream: reset by RST\n");
		}
		free_tcp(a_tcp);
		return ret;
	}
	
	/* The packet is ACKing data */
	if (this_tcphdr->flags.bits.ack)
	{
		if (from_client && a_tcp->client.state == TCP_SYN_SENT &&
		    a_tcp->server.state == TCP_SYN_RECV)
		{
			/* The ACK strikes back, part three of the TCP handshake trilogy */
			if (ntohl (this_tcphdr->ack) == a_tcp->server.seq)
			{
				a_tcp->client.state = TCP_ESTABLISHED;
				a_tcp->client.ack_seq = ntohl (this_tcphdr->ack);
				
				a_tcp->server.state=TCP_ESTABLISHED;
				a_tcp->nids_state=NIDS_JUST_EST;

				/* NOTIFY: Established */
				dprintf("tcpstream: established\n");

				a_tcp->nids_state = NIDS_DATA;
			}
		}
	}
	
	if (this_tcphdr->flags.bits.ack)
	{
		if ( ntohl(this_tcphdr->ack) > snd->ack_seq )
			snd->ack_seq=ntohl(this_tcphdr->ack);
		
		if (rcv->state == FIN_SENT)
			rcv->state = FIN_CONFIRMED;
		if (rcv->state == FIN_CONFIRMED
		    && snd->state == FIN_CONFIRMED)
		{
			a_tcp->nids_state = NIDS_CLOSE;
			return ret;
		}
	}
	
	/* Queue up actual packet data */
	if (datalen>0) {
		dprintf("tcpstream: data\n");
	}

	if ( this_tcphdr->flags.bits.fin ) 
		snd->state = FIN_SENT;

	/* Update the senders advertised window */
	snd->window = ntohs(this_tcphdr->win);
	
	return ret;
}

int tcp_init(void)
{
	int i;

	if ( !(tcp_stream_table=calloc(TCPHASH, sizeof (char *))) ) {
		return 0;
	}

	max_stream=64*TCPHASH;
	
	if ( !(streams_pool=malloc((max_stream+1) *
		sizeof (struct tcp_stream))) ) {
		free(tcp_stream_table);
		return 0;
	}
	
	printf("tcpstream: %u streams, %uKB\n",
		max_stream, (sizeof(struct tcp_stream)*max_stream)/1024);

	for (i = 0; i < max_stream; i++) {
		streams_pool[i].next_free = &(streams_pool[i + 1]);
	}

	streams_pool[max_stream].next_free = 0;
	free_streams = streams_pool;

	return 1;
}

void process_icmp (struct packet *p, unsigned int l)
{
	struct pkt_iphdr *iph=p->layer[l].h.ip;
	struct pkt_icmphdr *pkt=p->layer[l+1].h.icmp;
	struct pkt_iphdr *orig_ip;
	struct pkt_tcphdr *th;
	struct half_stream *hlf;
	int match_addr;
	struct tcp_stream *a_tcp;

	int from_client;
	int len = ntohs (iph->tot_len) - (iph->ihl << 2);

	if (len < sizeof (struct pkt_icmphdr))
		return;

	if (pkt->type != ICMP_DEST_UNREACH)
		return;

	len -= 8;

	if (len < sizeof (struct pkt_iphdr))
		return;

	orig_ip = (struct pkt_iphdr *) (((char *) pkt) + 8);
	len -= orig_ip->ihl << 2;
	if (len < 8)
		return;
	if ((pkt->code & 15) == ICMP_PROT_UNREACH ||
	    (pkt->code & 15) == ICMP_PORT_UNREACH)
		match_addr = 1;
	else
		match_addr = 0;
	if (pkt->code > NR_ICMP_UNREACH)
		return;
	if (match_addr && (iph->saddr != orig_ip->daddr))
		return;
	if (orig_ip->protocol != IPPROTO_TCP)
		return;
	th = (struct pkt_tcphdr *) (((char *) orig_ip) + (orig_ip->ihl << 2));
	if (!(a_tcp = find_stream (th, orig_ip, &from_client)))
		return;
	if (a_tcp->addr.dest == iph->daddr)
		hlf = &a_tcp->server;
	else
		hlf = &a_tcp->client;
	if (hlf->state != TCP_SYN_SENT && hlf->state != TCP_SYN_RECV)
		return;
	a_tcp->nids_state = NIDS_RESET;
	
	/* NOTIFY: RESET */
	dprintf("tcpstream: reset by ICMP\n");
	free_tcp (a_tcp);
}

/* Preprocessor entry point */
int tcpstream_process(struct packet *pkt, unsigned int i, void *priv)
{
	struct pkt_iphdr *iph=pkt->layer[i].h.ip;

	/* No layer decoded above IP */
	if ( pkt->llen < i+2 )
		return PREPROC_OK;

	/* Ignore fragments */
	if ( iph->frag_off & ipfmask )
		return PREPROC_OK;
	
	/* Ignore bad IP checksums - ooh naughty! */
	if ( !(pkt->layer[i].flags&FLAG_IP_CSUM) )
		return PREPROC_OK;

	/* Ignore broadcast and multicast packets */
	if ( (pkt->flags & (FP_MULTICAST|FP_BROADCAST)) )
		return PREPROC_OK;

#if 0
	/* Ignore truncated packets */
	if ( ((void *)iph+ntohs(iph->tot_len)) > pkt->end )
		return PREPROC_OK;
#endif	
	
	if (iph->protocol==6){
		pkt->layer[i].flags|=process_tcp(pkt, i)|FLAG_TCP_TRACK;
		if ( !(pkt->layer[i].flags&FLAG_TCP_STATE) ) {
			tcp_state_errors++;
		}tcp_packets++;
	}else if ( iph->protocol==1 ) {
		process_icmp(pkt, i);
	}

	return PREPROC_OK;
}

void tcpstream_free(void *priv)
{
	printf("tcpstream: %u state errors out of %u packets\n",
	       tcp_state_errors, tcp_packets);
}

int tcpstream_init(char *args)
{
	if ( tcpproto ) {
		printf("tcpstream: Can't add tcpstream twice!\n");
		return 0;
	}

	if ( !(tcpproto=decode_proto("tcp")) ) {
		printf("tcpstream: tcp protocol not loaded!\n");
		return 0;
	}

	if ( !tcp_init() ) {
		return 0;
	}
	
	/* TCP stream reassembly registering for IP realm?
	 * Well, we are only actually interested in TCP in IP */
	if (preproc_activate("ip", tcpstream_process, tcpstream_free, NULL)) {
		printf("tcpstream: TCP connection tracking ENABLED.\n");
		return 1;
	}else{
		return 0;
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_preproc_add preproc_add;
	int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("preproc.tcpstream", "TCP stream reassembly");
	PLUGIN_VERSION(0, 1);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(preproc_add=in->import("preproc.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_activate=in->import("preproc.activate")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=preproc_add("tcpstream", tcpstream_init);

	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
