/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*/
#include <config.h>
#include <stdio.h>
#include <plugin.h>

#ifdef HAVE_DLFCN_H
#include <dlfcn.h>
#endif

void *plugin_callback(const char *item)
{
	return NULL;
}

void firestat(char *an, char *fn)
{
	struct plugin_out pout;
	struct plugin_in pin;
	void *lib;
	proc_plugin_init init;
	proc_plugin_unload unload;

	if ( !(lib=dlopen(fn, RTLD_NOW)) ) {
		fprintf(stderr, "%s: %s\n", an, dlerror());
		return;
	}

	if ( !(init=dlsym(lib, PLUGIN_INIT)) ) {
		fprintf(stderr, "%s: %s: Couldn't resolve "
			PLUGIN_INIT "\n", an, fn);
		dlclose(lib);
		return;
	}

	if ( !(unload=dlsym(lib, PLUGIN_UNLOAD)) ) {
		fprintf(stderr, "%s: %s: Couldn't resolve "
			PLUGIN_UNLOAD "\n", an, fn);
		dlclose(lib);
		return;
	}

	pout.size=(sizeof(struct plugin_out));
	pin.size=(sizeof(struct plugin_in));
	pin.import=plugin_callback;

	init(&pin, &pout);

	printf("Filename: %s\n", fn);
	printf("Summary:  %s\n", pout.desc);
	printf("Name:     %s\n", pout.name);
	printf("Version:  %i.%i\n",
		pout.ver_major, pout.ver_minor);
	printf("Author:   %s <%s>\n",
		pout.author_name, pout.author_email);
	printf("License:  %s\n", pout.license);
	
	unload(0);
	
	dlclose(lib);
	
	/* If unload() prints messages group them with the
	 * right plugin... */
	printf("\n");
}

int main(int argc, char **argv)
{
	int i;

	for(i=1; i<argc; i++) {
		firestat(argv[0], argv[i]);
	}
	
	return 0;
}
